import os
import typing

from betterconf.caster import AbstractCaster
from betterconf.caster import DEFAULT_CASTER


class AbstractProvider:
    """Implement this class and pass to `field`"""

    def get(self, name: str) -> typing.Any:
        """Return a value or None"""
        raise NotImplementedError()


class EnvironmentProvider(AbstractProvider):
    """Default provider. Get vals from environment"""

    def get(self, name: str) -> typing.Any:
        return os.getenv(name)


DEFAULT_PROVIDER = EnvironmentProvider()


class Field:
    def __init__(
        self,
        name: str,
        default: typing.Optional[typing.Any] = None,
        provider: AbstractProvider = DEFAULT_PROVIDER,
        caster: AbstractCaster = DEFAULT_CASTER,
    ):
        self._value = provider.get(name)
        self._default = default
        self._caster = caster

    @property
    def value(self):
        if not self._value:
            if not self._default:
                raise ValueError("Variable is not found")
            return self._default
        return self._caster.cast(self._value)


class FieldInfo(typing.NamedTuple):
    name_to_set: str
    obj: Field


def field(
    name: str,
    default: typing.Optional[typing.Any] = None,
    provider: AbstractProvider = DEFAULT_PROVIDER,
    caster: AbstractCaster = DEFAULT_CASTER,
) -> Field:
    return Field(name, default, provider, caster)


def is_dunder(name: str) -> bool:
    if name.startswith("__") and name.endswith("__"):
        return True
    else:
        return False


def parse_objects(
    obj_to_parse: typing.Union[typing.Type["Config"], "Config"]
) -> typing.List[FieldInfo]:
    result = []
    for var in dir(obj_to_parse):
        name_to_set = var
        obj = getattr(obj_to_parse, var)
        if not isinstance(obj, Field):
            continue
        if is_dunder(name_to_set):
            continue
        result.append(FieldInfo(name_to_set=name_to_set, obj=obj))
    return result


class Config:
    def __init__(self, **to_override):
        result = parse_objects(self)
        for obj in result:
            if obj.name_to_set in to_override:
                setattr(self, obj.name_to_set, to_override.get(obj.name_to_set))
                continue
            setattr(self, obj.name_to_set, obj.obj.value)


__all__ = ("Config", "field", "AbstractProvider", "EnvironmentProvider")



import typing


class AbstractCaster:
    def cast(self, val: str):
        """Try to cast or return val"""
        raise NotImplementedError()


class BoolCaster(AbstractCaster):

    ABLE_TO_CAST = {"true": True, "false": False}

    def cast(self, val: str) -> typing.Union[str, bool]:
        as_bool = self.ABLE_TO_CAST.get(val.lower())
        if not as_bool:
            return val
        else:
            return as_bool


class IntCaster(AbstractCaster):
    def cast(self, val: str) -> typing.Union[str, int]:
        try:
            as_int = int(val)
            return as_int
        except ValueError:
            return val


class NothingCaster(AbstractCaster):
    """Caster who does nothing"""

    def cast(self, val: str) -> str:
        return val


to_bool = BoolCaster()
to_int = IntCaster()
DEFAULT_CASTER = NothingCaster()

__all__ = ("to_bool", "to_int", "AbstractCaster", "DEFAULT_CASTER")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['betterconf']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'betterconf',
    'version': '1.8',
    'description': 'Minimalistic Python library for your configs.',
    'long_description': '# Minimalistic Python library for your configs.\n\n## How to?\nAt first, install libary:\n\n```sh\npip install betterconf\n```\n\nAnd... write simple config:\n```python\nfrom betterconf import field, Config\n\nclass MyConfig(Config):\n    my_var = field("my_var")\n\ncfg = MyConfig()\nprint(cfg.my_var)\n```\n\nTry to run:\n```sh\nmy_var=1 python our_file.py\n```\n\nWith default values:\n```python\nfrom betterconf import field, Config\n\nclass MyConfig(Config):\n    my_var = field("my_var", default="hello world")\n\ncfg = MyConfig()\nprint(cfg.my_var)\n# hello world\n```\n\nOverride values when it\'s needed (for an example: test cases)\n```python\nfrom betterconf import field, Config\n\nclass MyConfig(Config):\n    my_var = field("my_var", default="hello world")\n\ncfg = MyConfig(my_var="WOW!")\nprint(cfg.my_var)\n# WOW!\n```\n\nBy default `betterconf` gets all values from `os.environ` but sometimes we need much.\nYou can create own `field value provider` in minutes:\n\n```python\nfrom betterconf import field, Config\nfrom betterconf.config import AbstractProvider\n\nclass NameProvider(AbstractProvider):\n    def get(self, name: str):\n        return name\n\nclass Cfg(Config):\n    my_var = field("my_var", provider=NameProvider())\n\ncfg = Cfg()\nprint(cfg.my_var)\n# my_var\n```\n\nAlso we can cast our values to python objects (or just manipulate them):\n\n```python\nfrom betterconf import field, Config\n# out of the box we have `to_bool` and `to_int`\nfrom betterconf.caster import to_bool, to_int, AbstractCaster\n\n\nclass DashToDotCaster(AbstractCaster):\n    def cast(self, val: str):\n        return val.replace("-", ".")\n\nto_dot = DashToDotCaster()\n\nclass Cfg(Config):\n    integer = field("int", caster=to_int)\n    boolean = field("bool", caster=to_bool)\n    dots = field("dashes", caster=to_dot)\n\ncfg = Cfg()\nprint(cfg.integer, cfg.boolean, cfg.dots)\n# -500, True, hello.world\n\n```\n\n```sh\ninteger=-500 boolean=true dots=hello-world python our_file.py\n```\n\n\n\n## License\nThis project is licensed under MIT License.\n\nSee [LICENSE](LICENSE) for details.\n\n\nMade with :heart: by [prostomarkeloff](https://github.com/prostomarkeloff) and our contributors.\n',
    'author': 'prostomarkeloff',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/prostomarkeloff/betterconf',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.1,<4.0',
}


setup(**setup_kwargs)
